/*
 * 2013-12-17 Sebastian Rockel (rockel@informatik.uni-hamburg.de)
 *
 * This script launches the RACE basic system in an automated, deterministic and
 * ordered way.  It depends on having all RACE components properly installed.
 * Furthermore it depends on having the blackboard locally installed on the PC
 * this is running on.
 *
 * It runs for the given amount of cycles and executes the given task. For
 * convenience real-time information as well as summaries after each runs
 * related to Fluent processing and time used is given on the terminal.
 */
#include "sim_manipulator.h"
#include "sim_creator.h"
#include "ros/package.h"
#include <stdio.h>
#include <boost/thread.hpp>

int main(int argc, char **argv)
{
  uint32_t simcount = 1;
  bool success = true;
  std::string scenario = "";

  if (argc >= 2)
  {
    scenario = argv[1];
  }
  else
  {
    scenario = "demo_21a";
  }
  scenario += ":=true";


  printf("[ INFO]: Starting %d simulation(s) in a row\n", simcount);
  printf("[ INFO]: Starting scenario with %s\n", scenario.c_str());

  // run N simulations
  for (uint32_t i = 0; i < simcount; ++i)
  {
    bool stop = false;
    std::stringstream model_file_content;
    SimCreator simCreator;
    std::vector<FluentSimObject> objects_vector;
    boost::thread_group group_th;
    volatile bool timer_elapsed = false;
    uint32_t timer_seconds_elapse = 3600; // scenario should be finished within 20 minutes
    double scenario_time = 0.0;
    int32_t number_fluents = 0;
    std::string scenario_instruction;
    std::string current_episode_name;

    printf("[ INFO]: Loading simulation run %d..\n", i+1);

    // Init a ROS node otherwise any creation of a ROS node will fail..
    printf("[ INFO]: Starting ROS..\n");
    group_th.create_thread(boost::bind(&system, "roslaunch race_gazebo_worlds race_world.launch furniture:=false gui:=true &>> /dev/null"));
    //group_th.create_thread(boost::bind(&system, "roslaunch race_gazebo_worlds race_world.launch furniture:=false gui:=false &>> /dev/null"));
    //group_th.create_thread(boost::bind(&system, "roslaunch race_gazebo_worlds race_world.launch furniture:=false gui:=false"));
    sleep(5);

    /********************************
     * Initialization of a ROS node *
     ********************************/

    ros::init(argc, argv, "sim_run"+i);
    ros::NodeHandle nh;

    // Uses ROS nodehandle, thus has to be called after (!) ros::init
    SimManipulator simManipulator(nh,
        tf::Transform(tf::Quaternion(0, 0, 0, 1), tf::Vector3(-12.28, -10.20, 0.0)));

    if (ros::ok())
    {
      simManipulator.setUpdateRate(1.0);
      simManipulator.unpause(); // Start simulation time
    }

    if (ros::ok())
    {
      printf("[ INFO] [%4.2f]: Spawn PR2..\n", ros::Time::now().toSec());
      group_th.create_thread(boost::bind(system, "roslaunch race_gazebo_worlds spawn_pr2.launch"));
      sleep(10); // wait for pr2 to come up
      ROS_INFO("Requesting model poses..");
      simManipulator.getModelPoses();

    }

    if (ros::ok())
    {
      printf("[ INFO] [%4.2f]: Spawning scenario items..\n", ros::Time::now().toSec());
      std::string scen_items = "roslaunch race_gazebo_worlds spawn_scenario_items.launch furniture:=true " +
        scenario + " &>> /dev/null";
      printf("[ INFO] [%4.2f]: %s..\n", ros::Time::now().toSec(), scen_items.c_str());
      group_th.create_thread(boost::bind(system, scen_items.c_str()));
    }

    /******************************************************
     * Simulation/PR2 systems should be running from here *
     ******************************************************/
    /* pretend this is a node
       listen on specific topics/use specific services to gain knowledge
       of current state of the RACE system
    */
    //ros::Rate loop_rate(1); // 1Hz

    if (ros::ok())
    {
      simManipulator.setUpdateRate(1.0);

      // Wait until SIGINT signal
      ROS_INFO("[ INFO]: Wait to stop ROS nodes..");
    }

    /*************************************************
     * This node's run loop. Do periodic tasks here. *
     *************************************************/
    while (ros::ok())
    {
      /* suggestion:
         usage of 'Timer' (see http://www.ros.org/wiki/roscpp/Overview/Timers)
         taken from http://answers.ros.org/question/11887/significance-of-rosspinonce/
      */
      /* loop_rate.sleep(); does not work, because something is wrong with sim-time and ros-time stuff!
         to see for yourself, comment in the section before, start the sim_run node
         after the blocking of the process just press 'play' in the gazebo simulator
         the process then proceeds to run but blocks forever after gazebo is killed!
         according to some rosanswers posts, this is related to usage of the '--clock' parameter
      */
      ros::spinOnce();
      sleep(1);

      // listen to specific topics (FAILURE SUCCESS)

      /* stop the system if something is broken,
        of if some criteria is met
        (i.e. failure or success of specific instruction)
        */
      //if (stop)
      //{
       //break;
      //}
    }
    /***********************************
     * ROS is dead after this line (!) *
     ***********************************/
    printf("[ WARN]: Stopping ROS nodes..\n");
    simCreator.stopRos();
    group_th.interrupt_all();
    printf("[ INFO]: ..stopped\n");

  }

  return 0;
}
